/*
 * Copyright (c) 2020. JetBrains s.r.o.
 * Use of this source code is governed by the MIT license that can be found in the LICENSE file.
 */

package jetbrains.datalore.plot.base.geom

import jetbrains.datalore.base.geometry.DoubleRectangle
import jetbrains.datalore.base.geometry.DoubleVector
import jetbrains.datalore.plot.base.*
import jetbrains.datalore.plot.base.aes.AesScaling
import jetbrains.datalore.plot.base.geom.legend.VLineLegendKeyElementFactory
import jetbrains.datalore.plot.base.geom.util.GeomHelper
import jetbrains.datalore.plot.base.geom.util.GeomUtil
import jetbrains.datalore.plot.base.geom.util.HintColorUtil
import jetbrains.datalore.plot.base.interact.GeomTargetCollector
import jetbrains.datalore.plot.base.render.LegendKeyElementFactory
import jetbrains.datalore.plot.base.render.SvgRoot
import jetbrains.datalore.vis.svg.SvgLineElement
import kotlin.math.max

class VLineGeom : GeomBase() {

    override val legendKeyElementFactory: LegendKeyElementFactory
        get() = LEGEND_KEY_ELEMENT_FACTORY

    override fun buildIntern(
        root: SvgRoot,
        aesthetics: Aesthetics,
        pos: PositionAdjustment,
        coord: CoordinateSystem,
        ctx: GeomContext
    ) {
        val geomHelper = GeomHelper(pos, coord, ctx)
        val helper = geomHelper.createSvgElementHelper()
        helper.setStrokeAlphaEnabled(true)

        val viewPort = when {
            ctx.flipped -> ctx.getAesBounds().flip()
            else -> ctx.getAesBounds()
        }
        val colorMarkerMapper = HintColorUtil.createColorMarkerMapper(GeomKind.V_LINE, ctx)

        val lines = ArrayList<SvgLineElement>()
        for (p in GeomUtil.withDefined(aesthetics.dataPoints(), Aes.XINTERCEPT)) {
            val intercept = p.interceptX()!!
            if (viewPort.xRange().contains(intercept)) {
                val start = DoubleVector(intercept, viewPort.top)
                val end = DoubleVector(intercept, viewPort.bottom)
                val line = helper.createLine(start, end, p)
                lines.add(line)

                val width = max(AesScaling.strokeWidth(p), 2.0) * 2.0
                val origin = DoubleVector(intercept - width / 2, end.y)
                val dimensions = DoubleVector(width, 0.0)
                val rect = DoubleRectangle(origin, dimensions)
                ctx.targetCollector.addRectangle(
                    p.index(),
                    geomHelper.toClient(rect, p),
                    GeomTargetCollector.TooltipParams(
                        markerColors = colorMarkerMapper(p)
                    )
                )
            }
        }

        lines.forEach { root.add(it) }
    }

    companion object {
        const val HANDLES_GROUPS = false
        val LEGEND_KEY_ELEMENT_FACTORY: LegendKeyElementFactory =
            VLineLegendKeyElementFactory()
    }
}
